import typing as tp


def bad_symbol(c: str) -> bool:
    return c.isspace() or c == '[' or c == '('


def get_borders(line: int, column: int, end_line: int, end_column: int, 
                file_content: tp.List[str]) -> tp.Tuple[int, int]:
    while bad_symbol(file_content[line - 1][column - 1]):
        line, column = inc_position(line, column, file_content)

    while bad_symbol(file_content[end_line - 1][end_column - 1]):
        end_line, end_column = dec_position(end_line, end_column, file_content)

    return get_offset(line, column, file_content), get_offset(end_line, end_column, file_content)


def get_offset(line: int, column: int, file_content: tp.List[str]) -> int:
    return sum([len(x) for x in file_content[:line-1]]) + column - 1


def inc_position(line: int, column: int, file_content: tp.List[str]) -> tp.Tuple[int, int]:
    if column == len(file_content[line - 1]):
        line += 1
        column = 1
    else:
        column += 1
    return line, column


def dec_position(line: int, column: int, file_content: tp.List[str]) -> tp.Tuple[int, int]:
    column -= 1
    if column == 0:
        line -= 1
        column = len(file_content[line - 1])
    return line, column



import typing as tp
import copy

from mypy.nodes import *
from mypy.traverser import *
import mypy.types

import utbot_mypy_runner.nodes as my_nodes


class MyTraverserVisitor(TraverserVisitor):
    def __init__(self, types, processor: tp.Callable[[int, int, int, int, mypy.types.Type, Any], None], definitions: tp.Dict[str, Any], annotation_node_dict: tp.Dict[str, Any], meta: my_nodes.Meta):
        self.types = types
        self.processor = processor
        self.meta: tp.Optional[my_nodes.Meta] = meta
        self.depth = 0
        self.definitions = definitions
        self.annotation_node_dict = annotation_node_dict

    def process_expression(self, o: Expression) -> None:
        if self.meta is not None and o in self.types.keys() and not isinstance(self.types[o], mypy.types.AnyType) \
            and o.end_line is not None and o.end_column is not None and o.line >= 0:
            self.processor(o.line, o.column, o.end_line, o.end_column, self.types[o], self.meta)

    def visit_name_expr(self, o: NameExpr) -> None:
        self.process_expression(o)
        super().visit_name_expr(o)

    def visit_member_expr(self, o: MemberExpr) -> None:
        self.process_expression(o)
        super().visit_member_expr(o)

    def visit_func(self, o: FuncItem) -> None:
        old_meta = self.meta
        definition: tp.Optional[my_nodes.Definition] = None
        
        containing_class: tp.Optional[my_nodes.AnnotationNode] = None
        if old_meta is not None and old_meta.containing_class is not None:
            containing_class = self.annotation_node_dict[old_meta.containing_class]
        
        if containing_class is not None and isinstance(containing_class, my_nodes.CompositeAnnotationNode):
            definition = next((x for x in containing_class.members if isinstance(x, my_nodes.FuncDef) and x.name == o.name), None)
        
        elif self.depth == 0 and o.name in self.definitions.keys():
            definition = self.definitions[o.name]

        if definition is not None and old_meta is not None:
            self.meta = copy.deepcopy(old_meta)
            self.meta.fullname_to_node_id[""] = definition.type.node_id
        else:
            self.meta = None

        self.depth += 1
        #print("META after func:", self.meta)
        #print("old meta:", self.meta)
        #print("name:", o.name)
        super().visit_func(o)
        self.depth -= 1
        self.meta = old_meta

    def visit_class_def(self, o: ClassDef) -> None:
        old_meta = self.meta
        self.meta = copy.copy(old_meta)

        if self.meta is not None and o.name in self.definitions.keys():
            self.meta.containing_class = self.definitions[o.name].type.node_id
        else:
            self.meta = None

        self.depth += 1
        super().visit_class_def(o)
        self.depth -= 1
        self.meta = old_meta


"""
    def visit_yield_expr(self, o: YieldExpr) -> None:
        self.process_expression(o)
        super().visit_yield_expr(o)

    def visit_call_expr(self, o: CallExpr) -> None:
        self.process_expression(o)
        super().visit_call_expr(o)

    def visit_op_expr(self, o: OpExpr) -> None:
        self.process_expression(o)
        super().visit_op_expr(o)

    def visit_comparison_expr(self, o: ComparisonExpr) -> None:
        self.process_expression(o)
        super().visit_comparison_expr(o)

    def visit_slice_expr(self, o: SliceExpr) -> None:
        self.process_expression(o)
        super().visit_slice_expr(o)

    def visit_cast_expr(self, o: CastExpr) -> None:
        self.process_expression(o)
        super().visit_cast_expr(o)

    def visit_assert_type_expr(self, o: AssertTypeExpr) -> None:
        self.process_expression(o)
        super().visit_assert_type_expr(o)

    def visit_reveal_expr(self, o: RevealExpr) -> None:
        self.process_expression(o)
        super().visit_reveal_expr(o)

    def visit_assignment_expr(self, o: AssignmentExpr) -> None:
        self.process_expression(o)
        super().visit_assignment_expr(o)

    def visit_unary_expr(self, o: UnaryExpr) -> None:
        self.process_expression(o)
        super().visit_unary_expr(o)

    def visit_list_expr(self, o: ListExpr) -> None:
        self.process_expression(o)
        super().visit_list_expr(o)

    def visit_tuple_expr(self, o: TupleExpr) -> None:
        self.process_expression(o)
        super().visit_tuple_expr(o)

    def visit_dict_expr(self, o: DictExpr) -> None:
        self.process_expression(o)
        super().visit_dict_expr(o)

    def visit_set_expr(self, o: SetExpr) -> None:
        self.process_expression(o)
        super().visit_set_expr(o)

    def visit_index_expr(self, o: IndexExpr) -> None:
        self.process_expression(o)
        super().visit_index_expr(o)

    def visit_generator_expr(self, o: GeneratorExpr) -> None:
        self.process_expression(o)
        super().visit_generator_expr(o)

    def visit_dictionary_comprehension(self, o: DictionaryComprehension) -> None:
        self.process_expression(o)
        super().visit_dictionary_comprehension(o)

    def visit_list_comprehension(self, o: ListComprehension) -> None:
        self.process_expression(o)
        super().visit_list_comprehension(o)

    def visit_set_comprehension(self, o: SetComprehension) -> None:
        self.process_expression(o)
        super().visit_set_comprehension(o)

    def visit_conditional_expr(self, o: ConditionalExpr) -> None: 
        self.process_expression(o)
        super().visit_conditional_expr(o)

    def visit_type_application(self, o: TypeApplication) -> None:
        self.process_expression(o)
        super().visit_type_application(o)

    def visit_lambda_expr(self, o: LambdaExpr) -> None:
        self.process_expression(o)
        super().visit_lambda_expr(o)

    def visit_star_expr(self, o: StarExpr) -> None:
        self.process_expression(o)
        super().visit_star_expr(o)

    def visit_backquote_expr(self, o: BackquoteExpr) -> None:
        self.process_expression(o)
        super().visit_backquote_expr(o)

    def visit_await_expr(self, o: AwaitExpr) -> None:
        self.process_expression(o)
        super().visit_await_expr(o)

    def visit_super_expr(self, o: SuperExpr) -> None:
        self.process_expression(o)
        super().visit_super_expr(o)
"""



import typing as tp
import mypy.nodes
import utbot_mypy_runner.mypy_main as mypy_main


class Name:
    def __init__(self, name, type_='Other'):
        self.name = name
        self.type_ = type_

    def encode(self):
        return {'name': self.name, 'kind': self.type_}


class ModuleName(Name):
    def __init__(self, name, fullname):
        super().__init__(name, 'Module')
        self.fullname = fullname
    
    def encode(self):
        superclass_dict = super().encode()
        subclass_dict = {'fullname': self.fullname}
        return dict(superclass_dict, **subclass_dict)


class LocalTypeName(Name):
    def __init__(self, name):
        super().__init__(name, 'LocalType')


class ImportedTypeName(Name):
    def __init__(self, name, fullname):
        super().__init__(name, 'ImportedType')
        self.fullname = fullname
    
    def encode(self):
        superclass_dict = super().encode()
        subclass_dict = {'fullname': self.fullname}
        return dict(superclass_dict, **subclass_dict)


def get_names_from_module(module_name: str, table: mypy.nodes.SymbolTable) -> tp.List[Name]:
    result: tp.List[Name] = []
    for name in table.keys():
        # TODO: remove synthetic names

        node = table[name].node

        if isinstance(node, mypy.nodes.TypeInfo):
            if node.is_intersection:
                continue

            if node._fullname.startswith(module_name):
                result.append(LocalTypeName(name))
            else:
                result.append(ImportedTypeName(name, node._fullname))

        elif isinstance(node, mypy.nodes.MypyFile):
            result.append(ModuleName(name, node._fullname))

        else:
            result.append(Name(name))

    return result


def get_names(build_result: mypy_main.build.BuildResult) -> tp.Dict[str, tp.List[Name]]:
    names_dict: tp.Dict[str, tp.List[Name]] = {}
    for module in build_result.files.keys():
        names_dict[module] = get_names_from_module(module, build_result.files[module].names)
    return names_dict


from mypy.main import *

import sys

from io import StringIO
from typing import List, Tuple, TextIO, Callable, Optional, cast


"""
Copy with some changes of function 'main' from here:
https://github.com/python/mypy/blob/v1.5.1/mypy/main.py
"""
def new_main(
    *,
    args: list[str] | None = None,
    stdout: TextIO = sys.stdout,
    stderr: TextIO = sys.stderr,
    clean_exit: bool = False,
) -> Optional[build.BuildResult]:
    """Main entry point to the type checker.

    Args:
        args: Custom command-line arguments.  If not given, sys.argv[1:] will
            be used.
        clean_exit: Don't hard kill the process on exit. This allows catching
            SystemExit.
    """
    util.check_python_version("mypy")
    t0 = time.time()
    # To log stat() calls: os.stat = stat_proxy
    sys.setrecursionlimit(2**14)
    if args is None:
        args = sys.argv[1:]

    fscache = FileSystemCache()
    sources, options = process_options(args, stdout=stdout, stderr=stderr, fscache=fscache)

    # CHANGE: export types of AST nodes
    options.preserve_asts = True
    options.export_types = True

    if clean_exit:
        options.fast_exit = False

    formatter = util.FancyFormatter(stdout, stderr, options.hide_error_codes)

    if options.install_types and (stdout is not sys.stdout or stderr is not sys.stderr):
        # Since --install-types performs user input, we want regular stdout and stderr.
        fail("error: --install-types not supported in this mode of running mypy", stderr, options)

    if options.non_interactive and not options.install_types:
        fail("error: --non-interactive is only supported with --install-types", stderr, options)

    if options.install_types and not options.incremental:
        fail(
            "error: --install-types not supported with incremental mode disabled", stderr, options
        )

    if options.install_types and options.python_executable is None:
        fail(
            "error: --install-types not supported without python executable or site packages",
            stderr,
            options,
        )

    # CHANGE
    # if options.install_types and not sources:
    #    install_types(formatter, options, non_interactive=options.non_interactive)
    #    return

    res, messages, blockers = run_build(sources, options, fscache, t0, stdout, stderr)

    if options.non_interactive:
        missing_pkgs = read_types_packages_to_install(options.cache_dir, after_run=True)
        if missing_pkgs:
            # Install missing type packages and rerun build.
            install_types(formatter, options, after_run=True, non_interactive=True)
            fscache.flush()
            print()
            res, messages, blockers = run_build(sources, options, fscache, t0, stdout, stderr)
        show_messages(messages, stderr, formatter, options)

    if MEM_PROFILE:
        from mypy.memprofile import print_memory_profile

        print_memory_profile()

    code = 0
    n_errors, n_notes, n_files = util.count_stats(messages)
    if messages and n_notes < len(messages):
        code = 2 if blockers else 1
    if options.error_summary:
        if n_errors:
            summary = formatter.format_error(
                n_errors, n_files, len(sources), blockers=blockers, use_color=options.color_output
            )
            stdout.write(summary + "\n")
        # Only notes should also output success
        elif not messages or n_notes == len(messages):
            stdout.write(formatter.format_success(len(sources), options.color_output) + "\n")
        stdout.flush()

    if options.install_types and not options.non_interactive:
        result = install_types(formatter, options, after_run=True, non_interactive=False)
        if result:
            print()
            print("note: Run mypy again for up-to-date results with installed types")
            code = 2

    return res


"""
Copy with some changes of mypy api functions from here:
https://github.com/python/mypy/blob/v0.971/mypy/api.py
"""
def _run(
    main_wrapper: Callable[[TextIO, TextIO], Optional[build.BuildResult]]
) -> Tuple[str, str, int, Optional[build.BuildResult]]:

    stdout = StringIO()
    stderr = StringIO()

    res = None
    try:
        res = main_wrapper(stdout, stderr)
        exit_status = 0
    except SystemExit as system_exit:
        exit_status = cast(int, system_exit.code)

    return stdout.getvalue(), stderr.getvalue(), exit_status, res


def run(args: List[str]) -> Tuple[str, str, int, Optional[build.BuildResult]]:
    args.append("--no-incremental")
    return _run(lambda stdout, stderr: new_main(args=args, stdout=stdout, stderr=stderr, clean_exit=True))


if __name__ == "__main__":
    import time
    start = time.time()
    stdout, stderr, exit_status, build_result = run(sys.argv[1:])
    print(f"Seconds passed: {time.time() - start}")
    print(stdout, stderr, exit_status, sep='\n')
    if build_result is None:
        print("BuildResult is None")
    else:
        pass
        #print(build_result.files['utbot_mypy_runner.nodes'].names['CompositeAnnotationNode'].node.names["module_key"].node.is_initialized_in_class)
        #print(build_result.files['builtins'].names['str'].node.names["count"].node.arguments[2].initializer)
        #print(build_result.files['dijkstra'].names['Dijkstra'].node.names['Node'].node.module_name)
        #print(build_result.files['numpy'].names['ndarray'].module_public)
        #print(build_result.files['_pytest.runner'].names['<subclass of "Node" and "tuple">'].node.is_intersection)
        #print(build_result.files['subtypes'].names['P'].node.names['f'].fullname)
        #print([build_result.files['subtypes'].names[x].fullname for x in build_result.files['subtypes'].names])
        #print(build_result.files['collections'].names['deque'].node.mro)
        #print(build_result.files['collections'].names['Counter'].node._promote)
        #for x in build_result.files['builtins'].names['set'].node.defn.type_vars:
        #    print(type(x))
